from op5helpers import *
import copy, time

standard_params = {
    "auth": ('monitor', 'monitor'),
    "verify": False
}
sections = [u"event"]

# Tip, for python noobs:
# If you try section("yoyo", {"content-type": "application/xml"}) you
# will notice a cryptic error message that means you should name your
# **extra arguments: section("yoyo", headers={"content-type": "application/xml"})

def section(name=None, **extra):
    """
    Helper for requesting a certain URI

    return requests.Response
    """
    params = copy.deepcopy(standard_params) # else we'd be updating by reference
    params.update(extra)
    url = api_endpoint + "api/report"
    if name:
        url += "/" + name
    r = requests.get(url, **params)
    if 'json' in r.headers.get('content-type', "").lower():
        try:
            _ = r.json()
        except ValueError:
            ok_(False, '(%d) Could not get JSON out of:\n%s' % (r.status_code, r.text))
    return r

log("Testing against: "+api_endpoint)

def test_no_std_auth_with_bad_credentials():
    r = section("event", auth=('p_diddys_daddy', 'puppylove'))
    eq_(r.status_code, 401, str(r.status_code)+": "+r.text)

def test_invalid_section():
    r = section("/jallah_akbar")
    eq_(r.status_code, 400, str(r.status_code)+": "+r.text)

def test_can_std_auth():
    r = section()
    eq_(r.status_code, 200, str(r.status_code)+": "+r.text)

def test_event_needs_options():
    r = section("event")
    eq_(r.json()["error"], "No options provided", str(r.status_code)+": "+r.text)
    eq_(r.status_code, 400, str(r.status_code)+": "+r.text)

@import_log("simple")
def test_event_exposed_properties():
    # This is a regression test. Additions and fortifications are obviously allowed, but removals should only happens
    # in major versions.
    container_keys = set([u'resource', u'start_time', u'limit', u'end_time', u'offset', u'generated_at', u'data', u'start_id', u'end_id'])
    alert_keys = set([u'retry', u'event_type', u'timestamp', u'hard', u'state', u'host_name', u'output', u'service_description', u'id'])
    downtime_keys = set([u'event_type', u'timestamp', u'host_name', u'service_description', u'id'])
    def proptest(params, nmatches, container_keys, custom_test = False):
        r = section("event", params=params)
        eq_(r.status_code, 200, str(r.status_code)+": "+r.text)
        eq_(len(r.json()['data']), nmatches, "Expected " + str(nmatches) + " lines of data in " + json.dumps(r.json()))
        container_key_diff =  container_keys ^ set(r.json().keys())
        eq_(container_key_diff, set(), "Uh oh, you broke the event API by not sending the same keys as specified, diff:\n" + str(container_key_diff) + "\n" + r.text)

        for row in r.json()['data']:
            if row['event_type'] in ('host_alert', 'service_alert'):
                event_key_diff = alert_keys ^ set(row.keys())
            else:
                event_key_diff = downtime_keys ^ set(row.keys())
            eq_(event_key_diff, set(), "Uh oh, you broke the event API by not sending the same keys as specified, diff:\n" + str(event_key_diff) + "\n" + r.text)

        if custom_test:
            custom_test(r)

    # all events within timestamps/ids
    yield proptest, {'start_time': 1364906100, 'end_time': 1364906105, "limit": 5}, 5, container_keys,
    yield proptest, {'start_id': 2, 'end_id': 7, "limit": 5}, 5, container_keys

    # subset of all events within timestamps/ids, check for 'next_resource'
    def extra(r):
        ok_('format=json' not in r.json()['next_resource'], 'Should not have included the requested format in the next_resource')
        ok_('format=json' not in r.json()['resource'], 'Should not have included the requested format in the resource')
    yield proptest, {'start_time': 1364906100, 'end_time': 1364906105, "limit": 1}, 1, container_keys | set(['next_resource']), extra
    yield proptest, {'start_id': 2, 'end_id': 7, "limit": 1}, 1, container_keys | set(['next_resource']), extra

    # include downtime
    def extra(r):
        ok_('include_downtime=1' in r.json()['resource'])
    yield proptest, {'start_time': 1364906100, 'end_time': 1364906107, 'include_downtime': True}, 8, container_keys, extra
    yield proptest, {'start_id': 2, 'end_id': 9, 'include_downtime': True}, 8, container_keys, extra


@import_log("simple")
def test_event_properties_input_translation():
    """ Make sure we can 'deselect' host alert types by specifying the
    string value 'service' to the option 'alert_types' """
    r = section('event', params={'start_time': 1364906100, 'end_time': 1364906105, 'alert_types': 'service'})
    eq_(r.status_code, 200, str(r.status_code)+": "+r.text)
    eq_(len(r.json()['data']), 3, "Expected 3 entries, got %i: %s" % (len(r.json()['data']), r.text))

@import_log("simple")
def test_event_properties_output_translation():
    r = section('event', params={'start_time': 1364906100, 'end_time': 1364906105})
    eq_(r.status_code, 200, str(r.status_code)+": "+r.text)
    eq_(len(r.json()['data']), 5, r.text)

    for entry in r.json()['data']:
        # cheap way of saying that "Nagios' properties are translated to strings"
        ok_(not is_number(entry['event_type']), "%s should not be a number" % entry['event_type'])
        ok_(not is_number(entry['state']), "%s should not be a number" % entry['state'])
        ok_(not entry.has_key('in_scheduled_downtime'), "in_scheduled_downtime should be gone")
        ok_(not entry.has_key('downtime_depth'), "in_scheduled_downtime should be gone")

@import_log("simple")
def test_event_time_pagination():
    first = section('event', params={'start_time': 1364906100, 'end_time': 1364906101, 'limit': 1})
    eq_(first.status_code, 200)

    eq_(len(first.json()['data']), 1, first.text)
    eq_(first.json()['data'][0]['output'], 'OK - laa-laa', first.json()['data'][0])
    # @todo add tests for valid url + valid parameters in the resource keys
    ok_('next_resource' in first.json(), first.text)
    eq_(first.json()['limit'], 1, first.text)
    eq_(first.json()['offset'], 0, first.text)
    eq_(first.json()['start_time'], str(1364906100), "expected start_time to be 1364906100, got " + json.dumps(first.json()))
    eq_(first.json()['end_time'], str(1364906101), first.text)
    ok_('generated_at' in first.json(), first.text)
    first_data = first.json()['data'][0]

    second = section('event', params={'start_time': 1364906100, 'end_time': 1364906101, 'limit': 1, 'offset': 1})
    eq_(second.status_code, 200, str(second.status_code)+": "+second.text)
    eq_(len(second.json()['data']), 1, second.text)
    eq_(second.json()['data'][0]['output'], 'OK - Sven Melander', second.json()['data'][0])
    eq_(second.json()['limit'], 1, first.text)
    eq_(second.json()['offset'], 1, first.text)
    eq_(second.json()['start_time'], str(1364906100), first.text)
    eq_(second.json()['end_time'], str(1364906101), first.text)

    second_data = second.json()['data'][0]
    ok_(cmp(first_data, second_data) != 0, "first: \n"+str(first_data)+"\n\nsecond:\n"+str(second_data))

def test_event_id_pagination():
    first = section('event', params={'start_id': 2, 'end_id': 5, 'limit': 1})
    eq_(first.status_code, 200)

    eq_(len(first.json()['data']), 1, first.text)
    eq_(first.json()['data'][0]['output'], 'OK - laa-laa', first.json()['data'][0])
    # @todo add tests for valid url + valid parameters in the resource keys
    ok_('next_resource' in first.json(), first.text)
    eq_(first.json()['limit'], 1, first.text)
    eq_(first.json()['offset'], 0, first.text)
    eq_(first.json()['start_id'], "2", "expected start_id to be 2, got " + json.dumps(first.json()))
    eq_(first.json()['end_id'], "5", first.text)
    ok_('generated_at' in first.json(), first.text)
    first_data = first.json()['data'][0]

    second = section('event', params={'start_id': 2, 'end_id': 5, 'limit': 1, 'offset': 1})
    eq_(second.status_code, 200, str(second.status_code)+": "+second.text)
    eq_(len(second.json()['data']), 1, second.text)
    eq_(second.json()['data'][0]['output'], 'OK - Sven Melander', second.json()['data'][0])
    eq_(second.json()['limit'], 1, first.text)
    eq_(second.json()['offset'], 1, first.text)
    eq_(second.json()['start_id'], "2", first.text)
    eq_(second.json()['end_id'], "5", first.text)

    second_data = second.json()['data'][0]
    ok_(cmp(first_data, second_data) != 0, "first: \n"+str(first_data)+"\n\nsecond:\n"+str(second_data))

@import_log("simple")
def test_event_traversing():
    """ Almost like the pagination test but a complete POC of how to traverse a possibly large result set """
    sys.path.insert(0, os.path.abspath("api/examples"))
    sys.path.insert(0, os.path.abspath("../api/examples"))
    try:
        import traverse
    except ImportError as e:
        log(e)
        assert False
    traverse.auth = standard_params['auth']
    traverse.params['report_period'] = 'custom'
    traverse.params['start_time'] = 1364906100
    traverse.api_server = api_endpoint
    result = traverse.traverse()
    ok_(len(result))

@import_log("simple")
def test_timeperiod():
    r = section('event', params={'report_period': 'last24hours', 'limit': 1})
    now = int(r.json()['generated_at'])
    before_last_24_hours = now - (24 * 60 * 60) - 1
    ok_(abs(before_last_24_hours - int(r.json()['start_time'])) < 20, "start time and 24 hours ago should not diff more than 20 seconds")
    ok_(abs(now - int(r.json()['end_time'])) < 20, "end time and now should not diff more than 20 seconds")

@import_log("simple")
def test_filters():
    def the_test(states, expected_objects):
        r = section('event', params=states)
        eq_(200, r.status_code, str(r.status_code) + ":\n" + r.text)
        data = r.json()['data']
        eq_(len(data), len(expected_objects), "Length of response data mismatch when querying\n%s\n\nGot length %d for %s:\n%s\n\nand length %d for %s:\n%s" % (r.url, len(data), "data", data, len(expected_objects), "expected_objects", expected_objects))
        for i in range(0, len(data)):
            eq_((data[i]['host_name'], data[i]['service_description']), expected_objects[i])
    yield (the_test, {'host_states': 'up', 'service_states': 'none'}, [(u'win-server1', u''), (u'printer1', u'')])
    yield (the_test, {'host_states[]': 'up', 'service_states': 'none'}, [(u'win-server1', u''), (u'printer1', u'')])
    yield (the_test, {'host_states[]': 'down', 'service_states': 'none'}, [(u'win-server1', u'')])
    yield (the_test, {'host_states[]': ['up', 'down'], 'service_states': 'none'}, [(u'win-server1', u''), (u'printer1', u''), (u'win-server1', u'')])
    yield (the_test, {'host_states[]': ['unreachable'], 'service_states': 'none'}, [])
    yield (the_test, {'host_states[]': ['none'], 'service_states': 'ok'}, [('win-server1', 'POP3 Server')])
    yield (the_test, {'host_states[]': ['up'], 'service_states': 'ok'}, [(u'win-server1', u''), (u'printer1', u''), ('win-server1', 'POP3 Server')])
    yield (the_test, {'host_states[]': ['up', 'down'], 'service_states[]': ['ok', 'warning']}, [(u'win-server1', u''), (u'printer1', u''), ('win-server1', 'POP3 Server'), ('win-server1', 'POP3 Server'), ('win-server1', 'POP3 Server'), (u'win-server1', u'')])
    yield (the_test, {'host_states[]': ['none'], 'service_states[]': ['none'], 'include_downtime': 1}, [(u'win-server1', u''), (u'win-server1', u'')])
    yield (the_test, {'host_states[]': ['up', 'down'], 'service_states[]': ['ok', 'warning'], 'include_downtime': 1}, [(u'win-server1', u''), (u'printer1', u''), ('win-server1', 'POP3 Server'), (u'win-server1', u''), ('win-server1', 'POP3 Server'), ('win-server1', 'POP3 Server'), (u'win-server1', u''), (u'win-server1', u'')])

def test_index_defaults_to_help():
    r = section(params={"format": "html"})
    eq_(r.status_code, 200)
    ok_("This is the REST-API" in r.content)

def test_help():
    r = requests.get(api_endpoint+"api/help/report", **standard_params)
    eq_(r.status_code, 200, str(r.status_code)+": "+r.text)
    r = requests.get(api_endpoint+"api/help/report/event", **standard_params)
    eq_(r.status_code, 200, str(r.status_code)+": "+r.text)
    r = requests.get(api_endpoint+"api/help/report/jsdahf", **standard_params)
    ok_(r.status_code != 200, str(r.status_code)+": "+r.text)

def test_content_negotiation():
    formats = {
        "xml": "application/xml; charset=utf-8",
        "json": "application/json; charset=utf-8",
        "html": "text/html; charset=utf-8"
    }
    for f in formats:
        content_type = formats[f]
        r = section("event", headers={'accept': content_type.split(";")[0]}, params={'limit': 1})
        eq_(r.headers["content-type"], content_type, str(r.status_code)+": "+r.text+"\n\nheaders:\n"+str(r.headers))

        r = section("event", params={"format": f})
        eq_(r.headers["content-type"], content_type, str(r.status_code)+": "+r.text+"\n\nheaders:\n"+str(r.headers))

def test_event_defaults_to_json():
    r = section("event")
    eq_(r.headers["content-type"], "application/json; charset=utf-8", str(r.status_code)+": "+r.text+"\n\nheaders:\n"+str(r.headers))

@import_log("simple")
def test_restrict_object_selection():
    params = {
        "start_time": 1364906100,
        "end_time": 1364906101,
        "host_name": ['printer1']
    }
    authed = section(
        "event",
        params=params
    )
    eq_(len(authed.json()['data']), 1, "Should have filtered out not-asked-for events:\n"+str(authed.json()))

def test_fails_without_params():
    r = section("event")
    eq_(r.status_code, 400)
    eq_(r.json()['error'], "No options provided")
    eq_(r.json()['full_error'], "No options provided")

# @todo test the report_api access right

@import_log("simple")
def test_limited_user():
    params = {
        "start_time": 1364906100,
        "end_time": 1364906106
    }
    authed = section(
        "event",
        params=params
    )
    group_name = 'shadonda'
    username = "latisha"

    ok_(authed.json() and 'data' in authed.json(), "No JSON?\n"+authed.text) # useful for debugging via var_dump in ninja
    eq_(len(authed.json()['data']), 6, "We should have the same amount of events as defined in simple.log (6), got " + str(len(authed.json()['data'])) + ":\n" + str(authed.json()))

    usergroup = Crud('usergroup', group_name)
    usergroup.post({
        'name': group_name,
        'access_rights': ['host_view_contact', 'api_report']
    })
    user = Crud('user', username)
    r = user.post({
        'username': username,
        'password': username,
        'modules': ['Default'],
        'groups': [group_name]
    }, cmp=False)
    contact = Crud('contact', username)
    contact.post({
        'contact_name': username,
        'alias': username
    }, cmp=False)
    host = Crud('host', 'printer1', clean=False)
    host.put({
        'host_name': 'printer1',
        'contacts': [username]
    })
    Change.post()

    not_authed = section(
        "event",
        params=params,
        auth=(username, username)
    )
    eq_(not_authed.status_code, 200, str(not_authed.status_code)+": "+not_authed.text)
    ok_(not_authed.json() and 'data' in not_authed.json(), "No JSON?\n"+not_authed.text) # useful for debugging via var_dump in ninja
    eq_(len(not_authed.json()['data']), 1, "We got " + str(len(authed.json()['data'])) + " events instead of only 1, as defined in simple.log. Response:\n" + json.dumps(authed.json()))
    Crud('user', username)
    Crud('usergroup', group_name)
    Crud('contact', username)
    Change.post()

@import_log("simple")
def test_limits_are_inclusive():
    def the_test(param_couple):
        r = section('event', params=param_couple)
        eq_(200, r.status_code, str(r.status_code) + ":\n" + r.text)
        eq_(1, len(r.json()['data']), "All events should be in the given timeframe:\n" + r.text)
    for param_couple in [
            {'start_time': 1364906099, 'end_time': 1364906100},
            {'start_time': 1364906100, 'end_time': 1364906100},
            {'start_time': 1364906104, 'end_time': 1364906105},
            {'start_id': 2, 'end_id': 2},
            ]:
            yield the_test, param_couple

def test_timestamp_validation():
    r = section('event', params={'start_time': 1364906100, 'end_time': 136490099})
    eq_(400, r.status_code, str(r.status_code) + ":\n" + r.text)

def test_id_validation():
    r = section('event', params={'start_id': 2, 'end_id': 1})
    eq_(400, r.status_code, str(r.status_code) + ":\n" + r.text)

@import_log("simple")
def test_limit_by_object_name():
    r = section('event', params={'start_time': 1364906101, 'host_name[]': ['printer1']})
    eq_(200, r.status_code, str(r.status_code) + ":\n" + r.text)
    evs = len(r.json()['data'])
    eq_(1, evs, "Got %d events, wanted 1:\n%s" % (evs, r.text))

    # let's try to make the API translate a string to an array as well
    r = section('event', params={'start_time': 1364906101, 'host_name': 'printer1'})
    eq_(200, r.status_code, str(r.status_code) + ":\n" + r.text)
    evs = len(r.json()['data'])
    eq_(1, evs, "Got %d events, wanted 1:\n%s" % (evs, r.text))

@import_log("simple")
def test_state():
    container_keys = set([u'resource', u'time', u'generated_at', u'data'])
    # I'd like to add a hard key here. However, ninja currently doesn't return it,
    # and the effect can be simulated with two requests, so we can add it later.
    row_keys = set([u'state', u'host_name', u'in_scheduled_downtime', u'monitor_is_running'])
    def the_test(params, nobjs, rowcb=False):
        r = section('state', params=params)
        eq_(200, r.status_code, str(r.status_code) + ":\n" + r.text)
        try:
            evs = len(r.json()['data'])
        except ValueError, e:
            ok_(False, "Invalid json returned, got: " + r.text)
        eq_(nobjs, evs, "Got %d events, wanted %d:\n%s" % (evs, nobjs, r.text))
        container_key_diff =  container_keys ^ set(r.json().keys())
        eq_(container_key_diff, set(), "Uh oh, you broke the state API by not sending the same container keys as specified, diff:\n" + str(container_key_diff) + "\n" + r.text)

        for row in r.json()['data']:
            row_key_diff = row_keys ^ set(row.keys())
            eq_(row_key_diff, set(), "Uh oh, you broke the state API by not sending the same data keys as specified, diff:\n" + str(row_key_diff) + "\n" + r.text)
            ok_(row['state'], "There should be a state, but it's empty")
            if rowcb:
                rowcb(row)
    yield the_test, {'time': 1364906102, 'host_name': 'printer1'}, 1, lambda x: eq_(x['state'], 'up')
    yield the_test, {'time': 1364906102, 'host_name[]': 'printer1'}, 1
    yield the_test, {'time': 1364906102, 'host_name[]': ['printer1'], 'state_types': 'hard'}, 1
    yield the_test, {'time': 1364906102, 'host_name[]': ['printer1', 'linux-server1'], 'state_types': 'hard'}, 2
    # the second before the change should be undetermined:
    yield the_test, {'time': 1364906100, 'host_name': 'printer1'}, 1, lambda x: eq_(x['state'], 'undetermined')
    # the second of the change should not apply:
    yield the_test, {'time': 1364906101, 'host_name': 'printer1'}, 1, lambda x: eq_(x['state'], 'undetermined')
    # the second after the change should apply
    yield the_test, {'time': 1364906102, 'host_name': 'printer1'}, 1, lambda x: eq_(x['state'], 'up')

    yield the_test, {'time': 1364906100, 'host_name': 'win-server1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 0)
    # downtime starts during this second, should not be included
    yield the_test, {'time': 1364906102, 'host_name': 'win-server1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 0)
    yield the_test, {'time': 1364906103, 'host_name': 'win-server1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 1)
    # downtime ends during this second, should not be included
    yield the_test, {'time': 1364906107, 'host_name': 'win-server1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 1)
    yield the_test, {'time': 1364906108, 'host_name': 'win-server1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 0)

    # other host not affected by win-server's downtime
    yield the_test, {'time': 1364906105, 'host_name': 'printer1'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 0)
    # host is in downtime, service should be reported as being, too
    row_keys |= set(['service_description'])
    yield the_test, {'time': 1364906105, 'service_description': 'win-server1;PING'}, 1, lambda x: eq_(x['in_scheduled_downtime'], 1)
